#include <nemo/stdint.h>
#include <nemo/unixtypes.h>

#include <nemo/stralloc.h>
#include <nemo/djbio.h>
#include <nemo/error.h>
#include <nemo/djbiofd.h>
#include <nemo/getln.h>
#include <nemo/exit.h>
#include <nemo/tai.h>
#include <nemo/caldate.h>
#include <nemo/caltime.h>
#include <nemo/leapsecs.h>
#include <nemo/unix.h>
#include <nemo/macro_unused.h>

#include "mess822.h"
#include "die.h"

static void out(const char *buf, unsigned int len)
{
  if (djbio_put(djbiofd_outsmall, buf, len) < 0) _exit(111);
}

static mess822_time t;
static struct tai seconds;

static sa_vector tokens = SA_VECTOR;

static stralloc line = STRALLOC;

static char buf[32];

static void doit(void)
{
  struct caltime local;
  unsigned int j;
  unsigned int state;
  stralloc *token;
  char ch;

  stralloc_replace(&line, '\0', '\n');

  t.known = 0;
  if (!mess822_token(&tokens, &line)) die_nomem();
  if (!mess822_when(&t, &line)) die_nomem();

  if (!t.known) {
    out("\t", 1);
  }
  else {
    caltime_tai(&t.ct, &seconds);
    caltime_local(&local, &seconds, 0, 0);
    out(buf, caltime_fmt_ctime(&local, buf));
    out("\n", 1);
  }
  out(" ", 1);

  state = 1;
  /* 1: start; 2: middle; 3: immediately after space; 4: after semicolon */

  for (j = 0; j < tokens.len; ++j) {
    token = &tokens.va[j];
    ch = token->s[0];
    stralloc_copyb(&line, token->s + 1, token->len -1);
    if (ch == '(') {
#ifdef notdef
      if (state == 3) {
        out("\n\t   ", 5);
      }
#endif
      out("(", 1);
      stralloc_replace(&line, '\n', '\0');
      out(line.s, line.len);
      out(")", 1);
      if (state & 1) {
        state = 2;
      }
    }
    else if (ch == '%') {
      if (state == 3) {
        if (stralloc_case_equals(token, "%from")
            || stralloc_case_equals(token, "%by")
            || stralloc_case_equals(token, "%for")
            || stralloc_case_equals(token, "%id")) {
          out("\n\t ", 3);
        }
      }
      stralloc_replace(&line, '\n', '\0');
      out(line.s, line.len);
      if (state & 1) {
        state = 2;
      }
    }
    else if (ch == ';') {
      if ((state == 2) || (state == 3)) {
        out("\n\t ", 3);
      }
      out(&ch, 1);
      state = 4;
    }
    else if ((ch == ' ') || (ch == '\t')) {
      if ((state != 1) && (state != 3)) {
        out(" ", 1);
      }
      if (state == 2) {
        state = 3;
      }
    }
    else {
      out(token->s, token->len);
      if (state & 1) {
        state = 2;
      }
    }
  }
  out("\n", 1);
}

static sa_vector received = SA_VECTOR;

static mess822_action a[] = {
/*  name        f  c  value      a  w  */
  { "received", 0, 0, &received, 0, &t },
  { 0,          0, 0, 0,         0, 0 }
};

int main(int argc __UNUSED__, char **argv __UNUSED__)
{
  unsigned int i;

  PROGRAM = argv[0];
  if (leapsecs_init() < 0) die_leapsecs();

  if (!mess822_slurp_headers(djbiofd_insmall, &line, a)) {
    if (errno == error_proto) die_proto();
    if (errno == error_nomem) die_nomem();
    die_read("input");
  }

  for (i = 0; i < received.len; i++) {
    if (!stralloc_copy(&line, &received.va[i])) die_nomem();
    stralloc_trim(&line, "\n", 1);
    doit();
  }
  djbio_flush(djbiofd_outsmall);
  _exit(0);
}
