#include "mess822.h"

unsigned int mess822_key_value_pairs(sa_vector *keys, sa_vector *values, const stralloc *in)
{
  static sa_vector tokens = SA_VECTOR;
  static stralloc name = STRALLOC;
  static stralloc value = STRALLOC;

  unsigned int pos;
  unsigned int state;
  stralloc *token;
  char token_type;

  if (!sa_vector_erase(keys)) return 0;
  if (!sa_vector_erase(values)) return 0;
  if (!stralloc_erase(&name)) return 0;
  if (!stralloc_erase(&value)) return 0;

  if (!mess822_token(&tokens, in)) return 0;

  /*          key = value ; key = value ; ... */
  /* states: 1   2 3     4 1   2 3     4 1... */

  state = 1;
  pos = 0;

  while (pos < tokens.len) {
    token = &tokens.va[pos++];
    token_type = token->s[0];

    if (token_type == '(') {
      continue;
    }

    if (token_type == ' ' || token_type == '\t') {  /* FWS */
      continue;
    }

    if (token_type == ';') {
      if (state == 1) return 1;  /* bail */
      if (state == 4 || state == 2 || state == 3) {  /* allow for null values */
        if (!sa_vector_append(keys, &name)) return 0;
        if (!sa_vector_append(values, &value)) return 0;
	if (!stralloc_erase(&name)) return 0;
	if (!stralloc_erase(&value)) return 0;
        state = 1;
        continue;
      }
      return 1;  /* bail */
    }

    if (token_type == '=') {
      if (state == 2) {
        state = 3;
        continue;
      }
      if (state == 4) {
        if (!stralloc_append(&value, "=")) return 0;
        continue;
      }
      return 1;  /* bail */
    }

    if (token_type == '%') {
      if (state == 1 || state == 2) {
        if (!stralloc_catb(&name, token->s + 1, token->len - 1)) return 0;
        stralloc_lower(&name);
        state = 2;
        continue;
      }
      if (state == 3 || state == 4) {
        if (!stralloc_catb(&value, token->s + 1, token->len - 1)) return 0;
        state = 4;
        continue;
      }
      return 1;  /* bail */
    }

    if (state == 4) {  /* '<', '>', ',', ':', '@', '.'  */
      if (!stralloc_append(&value, &token_type)) return 0;
    }

    /* should we bail? */
  }
  if (state == 4 || state == 2 || state == 3) {  /* allow for null values */
    if (!sa_vector_append(keys, &name)) return 0;
    if (!sa_vector_append(values, &value)) return 0;
  }

  return 1;
}
