#include <nemo/scan.h>
#include <nemo/str.h>

#include "mess822.h"

unsigned int mess822_when(mess822_time *out, const stralloc *in)
{
  static struct caltime ct;
  static sa_vector tokens = SA_VECTOR;
  static stralloc buffer = STRALLOC;

  unsigned int state;
  unsigned int pos;
  unsigned int flag_numeric;
  long number;
  char ch;
  unsigned int known;
  stralloc *token;

  if (!mess822_token(&tokens, in)) return 0;

  /* Thu,     29 Aug 1996 14 : 04 : 52 CDT   */
  /* states: 1  2   3    4  5 6  7 8  9   10 */

  state = 1;
  known = 0;
  pos = 0;

  /* optional day of week */
  if (tokens.len > 2) {
    token = &tokens.va[0];
    if (token->s[0] == '%' && token->len == 4) {  /* 3 char day of week */
      token = &tokens.va[1];
      if (token->s[0] == ',') {
        pos += 2;
      }
    }
  }

  while (pos < tokens.len) {
    token = &tokens.va[pos++];
    ch = token->s[0];

    if (ch == ':') {
      if (state == 5) {
        state = 6;
      }
      if (state == 7) {
        state = 8;
      }
    }

    if (ch == '%') {
      if (!stralloc_copyb(&buffer, token->s + 1, token->len - 1)) return 0;
      if (!stralloc_0(&buffer)) return 0;
      flag_numeric = !buffer.s[scan_long(buffer.s, &number)];
      switch (state) {
        case 1:
          if (!flag_numeric) return 1;
          ct.date.day = (int)number;
          break;
        case 2:
          stralloc_lower(&buffer);
          if (stralloc_equalb(&buffer, "jan", 4)) {
            ct.date.month = 1;
            break;
          }
          if (stralloc_equalb(&buffer, "feb", 4)) {
            ct.date.month = 2;
            break;
          }
          if (stralloc_equalb(&buffer, "mar", 4)) {
            ct.date.month = 3;
            break;
          }
          if (stralloc_equalb(&buffer, "apr", 4)) {
            ct.date.month = 4;
            break;
          }
          if (stralloc_equalb(&buffer, "may", 4)) {
            ct.date.month = 5;
            break;
          }
          if (stralloc_equalb(&buffer, "jun", 4)) {
            ct.date.month = 6;
            break;
          }
          if (stralloc_equalb(&buffer, "jul", 4)) {
            ct.date.month = 7;
            break;
          }
          if (stralloc_equalb(&buffer, "aug", 4)) {
            ct.date.month = 8;
            break;
          }
          if (stralloc_equalb(&buffer, "sep", 4)) {
            ct.date.month = 9;
            break;
          }
          if (stralloc_equalb(&buffer, "oct", 4)) {
            ct.date.month = 10;
            break;
          }
          if (stralloc_equalb(&buffer, "nov", 4)) {
            ct.date.month = 11;
            break;
          }
          if (stralloc_equalb(&buffer, "dec", 4)) {
            ct.date.month = 12;
            break;
          }
          return 1;
        case 3:
          if (!flag_numeric) return 1;
          if (number < 50) {
            number += 2000;
          }
          if (number < 999) {
            number += 1900;
          }
          ct.date.year = number;
          break;
        case 4:
          if (!flag_numeric) return 1;
          ct.hour = (int)number;
          break;
        case 5:
          return 1;
        case 6:
          if (!flag_numeric) return 1;
          ct.minute = (int)number;
          break;
        case 8:
          if (!flag_numeric) return 1;
          ct.second = (int)number;
          break;
        case 7:
          ct.second = 0;
          state = 9;
        case 9:
          known = 2;
          if (flag_numeric) {  /* happiness */
            if (number >= 0) {
              ct.offset = (number / 100) * 60 + (number % 100);
            }
            else {
              ct.offset = -(((-number) / 100) * 60 + ((-number) % 100));
            }
            if (stralloc_case_equals(&buffer, "-0000")) known = 1;
            break;
          }
          stralloc_lower(&buffer);
          if (stralloc_equalb(&buffer, "ut", 3)) {
            ct.offset = 0;
            break;
          }
          if (stralloc_equalb(&buffer, "gmt", 4)) {
            ct.offset = 0;
            break;
          }
          if (stralloc_equalb(&buffer, "utc", 4)) {
            ct.offset = 0;
            break;
          }
          /* XXX: GMT+nnnn? */

          if (stralloc_equalb(&buffer, "bst", 4)) {
            ct.offset = 60;
            break;
          }
          if (stralloc_equalb(&buffer, "cdt", 4)) {
            ct.offset = -300;
            break;
          }
          if (stralloc_equalb(&buffer, "cet", 4)) {
            ct.offset = 60;
            break;
          }
          if (stralloc_equalb(&buffer, "cst", 4)) {
            ct.offset = -360;
            break;
          }
          if (stralloc_equalb(&buffer, "edt", 4)) {
            ct.offset = -240;
            break;
          }
          if (stralloc_equalb(&buffer, "eet", 4)) {
            ct.offset = 120;
            break;
          }
          if (stralloc_equalb(&buffer, "est", 4)) {
            ct.offset = -300;
            break;
          }
          if (stralloc_equalb(&buffer, "hkt", 4)) {
            ct.offset = 480;
            break;
          }
          if (stralloc_equalb(&buffer, "ist", 4)) {
            ct.offset = 120;
            break;
          }
          if (stralloc_equalb(&buffer, "jst", 4)) {
            ct.offset = 540;
            break;
          }
          if (stralloc_equalb(&buffer, "mdt", 4)) {
	    ct.offset = -360;
	    break;
          }
          if (stralloc_equalb(&buffer, "met", 4)) {
            ct.offset = 60;
            break;
          }
          if (stralloc_equalb(&buffer, "metdst", 7)) {
            ct.offset = 120;
            break;
          }
          if (stralloc_equalb(&buffer, "mst", 4)) {
            ct.offset = -420;
            break;
          }
          if (stralloc_equalb(&buffer, "pdt", 4)) {
            ct.offset = -420;
            break;
          }
          if (stralloc_equalb(&buffer, "pst", 4)) {
            ct.offset = -480;
            break;
          }
          return 1;
        case 10:
          stralloc_lower(&buffer);
          if (stralloc_equalb(&buffer, "dst", 4)) {
            ct.offset += 60;
          }
          break;
        default:
          break;
      }
      if (state < 10) {
        ++state;
      }
    }
  }

  if (known) {
    out->known = known;
    out->ct = ct;
  }
  return 1;
}
