#include <nemo/stralloc.h>
#include <nemo/str.h>
#include <nemo/byte.h>

#include "mess822.h"
#include "rewritehost.h"

static stralloc work = STRALLOC;

static unsigned int doit(const stralloc *rule)
{
  static stralloc new = STRALLOC;
  static stralloc post = STRALLOC;
  unsigned int colon;
  unsigned int prefixlen;
  char ch;

  if (!rule->len) return 1;
  ch = rule->s[0];
  if ((ch != '?') && (ch != '=') && (ch != '*') && (ch != '-')) return 1;
  colon = stralloc_chr(rule, ':');
  if (colon == rule->len) return 1;
  if (!stralloc_copyb(&post, rule->s + 1, colon - 1)) return 1;
  if (!stralloc_copyb(&new, rule->s + colon + 1, rule->len - colon - 1)) return 1;

  if (ch == '=') {
    if (stralloc_case_diff(&work, &post)) return 1;
    work.len = 0;
  }
  else {
    if (!stralloc_case_end(&work, &post)) return 1;
    if (ch == '*') {
      work.len -= post.len;
    }
    else if (ch == '?') {
      prefixlen = work.len - post.len;
      if (byte_chr(work.s, prefixlen, '.') < prefixlen) return 1;
      if (byte_chr(work.s, prefixlen, '[') < prefixlen) return 1;
      if (byte_chr(work.s, prefixlen, ']') < prefixlen) return 1;
      work.len -= post.len;
    }
    else {  /* ch == '-' */
      work.len = 0;
    }
  }

  return stralloc_cat(&work, &new);
}

static unsigned int append_work(stralloc *out, const sa_vector *rules)
{
  unsigned int j;

  for (j = 0; j < rules->len; ++j) {
    if (!doit(&rules->va[j])) return 0;
  }
  return stralloc_cat(out, &work);
}

static unsigned int append_addr(stralloc *out, const stralloc *in, const sa_vector *rules)
{
  unsigned int at;

  at = stralloc_chr(in, '@');
  if (!at) {
    if (in->len <= 1) return 1;
  }
  if (!stralloc_catb(out, in->s, at)) return 0;
  if (!stralloc_append(out, "@")) return 0;
  if (at < in->len) {
    ++at;
  }
  if (!stralloc_copyb(&work, in->s + at, in->len - at)) return 0;
  return append_work(out, rules);
}

unsigned int rewritehost(stralloc *out, const stralloc *in, const sa_vector *rules)
{
  if (!stralloc_erase(out)) return 0;
  if (!stralloc_copy(&work, in)) return 0;
  return append_work(out, rules);
}

unsigned int rewritehost_addr(stralloc *out, const stralloc *in, const sa_vector *rules)
{
  if (!stralloc_erase(out)) return 0;
  return append_addr(out, in, rules);
}

unsigned int rewritehost_list(mess822_addrlist *out, const mess822_addrlist *in, const sa_vector *rules)
{
  static stralloc data = STRALLOC;
  mess822_addrdata *item;
  unsigned int len;
  unsigned int j;

  if (!mess822_addrlist_erase(out)) return 0;
  len = in->len;
  for (j = 0; j < len; ++j) {
    item = &in->ad[j];
    if (item->type == '@') {
      if (!stralloc_erase(&data)) return 0;
      if (!append_addr(&data, &item->data, rules)) return 0;
    }
    else {
      if (!stralloc_copy(&data, &item->data)) return 0;
    }
    if (!mess822_addrlist_append(out, &data, item->type)) return 0;
  }
  return 1;
}
